
#ifndef NO_RTEMS
  // Include the RTEMS board specification
  #include <bsp.h>
  #include <rtems/untar.h>
  #include <rtems.h>
#else
  #include <string.h>
#endif  

#include <stdio.h>
#include <stdlib.h>

// Include the CDc code
#include "Constants.h"
#include "config.h"
#include "filesystem.h"
#include "utils.h"
#include "rapita.h"
// -------------------------------------------------------------------------------------------------
// Supporting functions
// -------------------------------------------------------------------------------------------------
inline char* copyString(char* from) {
    char* result=(char*)checkedMalloc(strlen(from)+1);
    return strcpy(result,from);
}

#ifndef NO_RTEMS

 cdc_clock clock_fromTimespec(struct timespec from) {
    // ugly patch to be on NS as java is
    long long r=(long long)(from.tv_nsec)/1000+(long long)(from.tv_sec)*1000000;
    printf("%llu\n",r);
    return r;
}

#else

inline cdc_clock clock_fromTimeval(struct timeval from) {
    return (from.tv_usec + from.tv_sec * 1000000);
}
#endif

// TODO I believe this function is not needed
cdc_clock clock_roundUp(cdc_clock from) {
    if (from%1000!=0) from+=(from%1000); // round-up to millisecond
    if (DETECTOR_PERIOD>0) {
        from=((from+DETECTOR_PERIOD*1000-1)/(DETECTOR_PERIOD*1000))*(DETECTOR_PERIOD*1000);
    }
    return from;
}



#ifdef NO_RTEMS

volatile long long __time;
struct timeval _clk;
inline cdc_clock clock_now() {
    gettimeofday(&_clk, NULL);
    return clock_fromTimeval(_clk);
}

#else

volatile long long __time;
struct timespec _clk;
void clock_now() {
    rtems_clock_get_uptime(&_clk); // FIXME: Is this the same precision as clock_gettime(CLOCK_REALTIME,&_clk) ?
    __time=(long long)(_clk.tv_nsec)/1000+(long long)(_clk.tv_sec)*1000000;
    return 0;
    return clock_fromTimespec(_clk);
}

#endif

int clock_roundUpTicks() {
    clock_now();
    cdc_clock c=__time;
    c=(500000-c%500000)/1000;
    return c;
}


// -------------------------------------------------------------------------------------------------
// CDc Arguments Generator
// -------------------------------------------------------------------------------------------------

#ifdef HARDCODED_ARGUMENTS
//char *defaultArgv[] = { CDC_PROGRAM_NAME, "mono10.bin", "MAX_FRAMES", "100", "DUMP_SENT_FRAMES", "DUMP_RECEIVED_FRAMES", "NUMBER_OF" };
char* defaultArgv[] ={ CDC_PROGRAM_NAME, "frames.bin", "MAX_FRAMES", "1000", "PERIOD", "120", "REDUCER_GRID_SIZE", "10"};
int defaultArgc = 8;
#endif

int _argcc=0;

#ifdef STDIN_ARGUMENTS /* arguments from standard input */
// -------------------------------------------------------------------------------------------------

#ifdef HARDCODED_ARGUMENTS
  #error "Both STDIN_ARGUMENTS and HARDCODED_ARGUMENTS defined"
#endif
  
/** \brief Reads new argument from the standard input and returns it.
 *
 * The result is owned by the function. If no other argument is present, returns NULL. First call
 * initializes the remote connection - awaits the handshake, then reads number of parameters and
 * returns the first parameter which is not read from the input, but is always "cdcrtems".
 */
#define CMD_BUFFER_LEN 4096
char _buffer[CMD_BUFFER_LEN];
int  _argc=-1;
//int  _argcc=0; // counter
int  _pos;
char* getNextArgument() {
    // Not initialized, perform handshake and
    if (_argc==-1) {
        printf("CLIENT_READY\n"); // was OVM something ready...
        while (1) {
            debug("Awaiting remote handshake");
            fgets(_buffer, CMD_BUFFER_LEN,stdin);
            if (!strncmp(_buffer,"CMDLINE-FOLLOWS",15)) {
                // ok handshake done, now read number of parameters
                debug("Number of arguments:");
                fgets(_buffer,CMD_BUFFER_LEN,stdin);
                _argc=atoi(_buffer);
                _argcc=0;
                return CDC_PROGRAM_NAME; // return program name, which in our case is cdc-rtems
            }
        }
        debug("Enter CMDLINE-FOLLOWS to continue");
    } else {
        // read a single argument and return it, or return NULL if no arguments are present
        if (_argcc>=_argc) return NULL;
        debug("Enter argument:");
        fgets(_buffer,CMD_BUFFER_LEN,stdin);
        if (strlen(_buffer)>1) {
            _buffer[strlen(_buffer)-1]=0;
            _argcc++;
        } else {
            fatalError("Argument not finished with a new line.");
        }
        return _buffer;
    }
}

#else /* hard-coded or real "main function" arguments */

char **effectiveArgv;
int effectiveArgc;

//int _argcc=0;
char* getNextArgument() {
    if (_argcc>=effectiveArgc) {
        return NULL;
    }
    return effectiveArgv[_argcc++];
}
#endif


// -------------------------------------------------------------------------------------------------
// CDc Application
// -------------------------------------------------------------------------------------------------
char* program_name;
char* input_file;
// -------------------------------------------------------------------------------------------------
/** \brief Displays application help.
 * usage -- tell the user how to use this program and   *
 *              exit                                    *
 */
void usage(void) {
    fprintf(stderr, "Usage is %s air_traffic.bin [options]\n" 
        "  PERIOD <num>             Period of the detector in milliseconds.\n"
        "  DETECTOR_PRIORITY <num>  Priority of the detector thread.\n"
        "  MAX_FRAMES <num>         Number of frames to process.\n"
        "  BUFFER_FRAMES <num>      Size of internal frame buffer. \n"
        "                           Must be at least MAX_FRAMES.\n"
        "  DEBUG_DETECTOR           Turns on some debugging messages in \n"
        "                           collision detection.\n"
        "  DUMP_RECEIVED_FRAMES     For debugging only. Dumps frames received\n"
        "                           by detector.\n"
        "  DUMP_SENT_FRAMES         For debugging only. Dumps frames sent to\n"
        "                           frame buffer at initialization.\n"
        "  REDUCER_GRID_SIZE        Size of the grid used by the reducer.\n"
        "\n",
        
            program_name);
    exit(0);
}

void dump_parameters(void) {

  fprintf(stderr, "Execution parameters: \n"
    " PERIOD = %d\n DETECTOR_PRIORITY = %d\n MAX_FRAMES = %d\n"
    " BUFFER_FRAMES = %d\n DEBUG_DETECTOR = %d\n DUMP_RECEIVED_FRAMES = %d\n"
    " DUMP_SENT_FRAMES = %d\n REDUCER_GRID_SIZE = %f\n\n",
        DETECTOR_PERIOD, DETECTOR_PRIORITY, MAX_FRAMES, BUFFER_FRAMES, 
        DEBUG_DETECTOR, DUMP_RECEIVED_FRAMES, DUMP_SENT_FRAMES,GOOD_VOXEL_SIZE);

}

char *getNextRequiredArgument() {
    char *arg = getNextArgument();
    
    if (!arg) {
        fprintf(stderr, "Too few command line arguments.\n");
        usage();
    }
    
    return arg;    
}

// -------------------------------------------------------------------------------------------------
/** \brief Incremental parser of the command line arguments.
 *
 * Uses the getNextArgument() as a generator. Parses all arguments given by this function in the
 * same way as CDc old parse function does.
 */
// TODO remove unused constants
void parseIncremental() {
    char* arg;
    while ((arg=getNextArgument())!=NULL) {
        if (0) {
//        } else if (strcmp("PERSISTENT_DETECTOR_SCOPE_SIZE", arg) == 0) {
//            PERSISTENT_DETECTOR_SCOPE_SIZE = atol(getNextArgument());
        } else if (strcmp("PERIOD", arg) == 0) {
            DETECTOR_PERIOD = atol(getNextRequiredArgument());
            // Rescale to RTEMS ticks
            // CHECK This is not 100% accurate if the number of ticks per millisecond is not
            // an integer, but I believe this would be (a) very rare and (b) imprecise even with
            // floating points as the detector period is long.
            DETECTOR_PERIOD = DETECTOR_PERIOD;
//        } else if (strcmp("TRANSIENT_DETECTOR_SCOPE_SIZE",arg) == 0) {
//            TRANSIENT_DETECTOR_SCOPE_SIZE = atol(getNextRequiredArgument());
        } else if (strcmp("DETECTOR_PRIORITY", arg) == 0) {
            DETECTOR_PRIORITY = atoi(getNextRequiredArgument());
//            DETECTOR_STARTUP_PRIORITY = DETECTOR_PRIORITY - 1;
//        } else if (strcmp("SIMULATOR_PRIORITY", arg) == 0) {
//            SIMULATOR_PRIORITY = atoi(getNextRequiredArgument());
//        } else if (strcmp("NOISE_RATE", arg) == 0) {
//            NOISE_RATE = atoi(getNextRequiredArgument());
        } else if (strcmp("MAX_FRAMES", arg) == 0) {
            MAX_FRAMES = atoi(getNextRequiredArgument());
//        } else if (strcmp("TIME_SCALE", arg) == 0) {
//            TIME_SCALE = atoi(getNextRequiredArgument());
        } else if (strcmp("BUFFER_FRAMES", arg) == 0) {
            BUFFER_FRAMES = atoi(getNextRequiredArgument());
        } else if (strcmp("REDUCER_GRID_SIZE", arg)==0) {
            GOOD_VOXEL_SIZE = atof(getNextRequiredArgument());

//        } else if (strcmp("FPS", arg) == 0) {
//            FPS = atoi(getNextRequiredArgument());
//        } else if (strcmp("DETECTOR_NOISE_REACHABLE_POINTERS", arg) == 0) {
//            DETECTOR_NOISE_REACHABLE_POINTERS = atoi(getNextRequiredArgument());
//        } else if (strcmp("DETECTOR_NOISE_ALLOCATION_SIZE", arg) == 0) {
//            DETECTOR_NOISE_ALLOCATION_SIZE = atoi(getNextRequiredArgument());
//        } else if (strcmp("DETECTOR_NOISE_ALLOCATE_POINTERS", arg) == 0) {
//            DETECTOR_NOISE_ALLOCATE_POINTERS = atoi(getNextRequiredArgument());
//        } else if (strcmp("DETECTOR_NOISE_MIN_ALLOCATION_SIZE", arg) == 0) {
//            DETECTOR_NOISE_MIN_ALLOCATION_SIZE = atoi(getNextRequiredArgument());
//        } else if (strcmp("DETECTOR_NOISE_MAX_ALLOCATION_SIZE", arg) == 0) {
//            DETECTOR_NOISE_MAX_ALLOCATION_SIZE = atoi(getNextRequiredArgument());
//        } else if (strcmp("DETECTOR_NOISE_ALLOCATION_SIZE_INCREMENT", arg) == 0) {
//            DETECTOR_NOISE_ALLOCATION_SIZE_INCREMENT = atoi(getNextRequiredArgument());
//        } else if (strcmp("DETECTOR_STARTUP_OFFSET_MILLIS", arg) == 0) {
//            DETECTOR_STARTUP_OFFSET_MILLIS = atoi(getNextRequiredArgument());
//        } else if (strcmp("SPEC_NOISE_ARGS", arg) == 0) {
//            SPEC_NOISE_ARGS = copyString(getNextRequiredArgument());
//        } else if (strcmp("SYNCHRONOUS_DETECTOR", arg) == 0) { 
                 
                 /*
		 * flags without
		 * a parameter
		 */
		 
//            SYNCHRONOUS_DETECTOR = 1;
//        } else if (strcmp("PRESIMULATE", arg) == 0) {
//            PRESIMULATE = 1;
//        } else if (strcmp("FRAMES_BINARY_DUMP", arg) == 0) {
//            FRAMES_BINARY_DUMP = 1;
        } else if (strcmp("DEBUG_DETECTOR", arg) == 0) {
            DEBUG_DETECTOR = 1;
        } else if (strcmp("DUMP_RECEIVED_FRAMES", arg) == 0) {
            DUMP_RECEIVED_FRAMES = 1;
        } else if (strcmp("DUMP_SENT_FRAMES", arg) == 0) {
            DUMP_SENT_FRAMES = 1;
//        } else if (strcmp("USE_SPEC_NOISE", arg) == 0) {
//            USE_SPEC_NOISE = 1;
//        } else if (strcmp("SIMULATE_ONLY", arg) == 0) {
//            SIMULATE_ONLY = 1;
//        } else if (strcmp("DETECTOR_NOISE", arg) == 0) {
//            DETECTOR_NOISE = 1;
//        } else if (strcmp("DETECTOR_NOISE_VARIABLE_ALLOCATION_SIZE", arg) == 0) {
//            DETECTOR_NOISE_VARIABLE_ALLOCATION_SIZE = 1;
        } else usage(); // Wrong argument
    }
    
    if (MAX_FRAMES > BUFFER_FRAMES) {
        fatalError("MAX_FRAMES is larger than BUFFER_FRAMES");
    }
}

// -------------------------------------------------------------------------------------------------
/** \brief Main application task.
 * 
 * Loads the workload and executes the detector.
 */
 
void the_main(void);

#ifndef NO_RTEMS
rtems_task cdc_task_main(rtems_task_argument ignored) {

#ifdef HARDCODED_ARGUMENTS
    effectiveArgc = defaultArgc;
    effectiveArgv = defaultArgv;
#else 
  #ifndef STDIN_ARGUMENTS
      error("RTEMS cannot use standard command line arguments.")
  #endif
#endif
      
    the_main();
    exit(0);
}
#else
int main(int argc, char **argv) {

#ifdef HARDCODED_ARGUMENTS
    effectiveArgc = defaultArgc;
    effectiveArgv = defaultArgv;
#else
    #ifndef STDIN_ARGUMENTS
        effectiveArgc = argc;
        effectiveArgv = argv;
    #endif
#endif

    the_main();
    exit(0);
}
#endif


extern int _ri;
void the_main(void) {

    // ---------------------------------------------------------------------------------------------
    // Insert test code here
    // ---------------------------------------------------------------------------------------------

    // ---------------------------------------------------------------------------------------------
    // End of test code
    // ---------------------------------------------------------------------------------------------
    benchmarkRapita(rapita_initialize());
    benchmarkRapita(rapita_set(RAPITA_BENCHMARK));
#ifndef NO_RTEMS
    // load the filesystem at first
    load_filesystem_image();
#endif    
    debug("CDc : initializing");
    // Set the program name and input file
    program_name = copyString(getNextArgument());
    
    char *_input_file = getNextArgument();
    if (!_input_file) {
        fprintf(stderr, "Too few command line arguments: missing file with air traffic dump.\n");
        exit(1);
    }
    input_file = copyString(_input_file);
    
    // parse input arguments
    parseIncremental();
    dump_parameters();
    // printf("Running with java threads and java memory\n"); not so much
    debug("CDc : initialized, starting...");
    IMMORTALENTRY_run();
    debug("CDc : reading input dump and then running detector...");
    /* read input dump */
    generate(input_file);
    debug("CDc : cleaning up...");
    /* clean-up */
    DETECTOR_finish();
    free(program_name);
    free(input_file);
    // ok, we are done
    debug("CDc : iteration done.");
    benchmarkRapita(rapita_set(RAPITA_DONE));
    debugp("Total rapita marks %u",_ri);
}

#ifndef NO_RTEMS

// -------------------------------------------------------------------------------------------------
// CDc Initialization Task
// -------------------------------------------------------------------------------------------------
// The CDc main task
rtems_id cdc_task;
/** \brief RTEMS Initialization task.
 *
 * Initializes the system, starts the main task and shuts itself down.
 */
rtems_task Init(rtems_task_argument ignored) {
    // create the task
    if (rtems_task_create(
            rtems_build_name('C','D','c',' '),
            DETECTOR_PRIORITY,
            RTEMS_MINIMUM_STACK_SIZE,
            RTEMS_NO_PREEMPT,
            (RTEMS_LOCAL | RTEMS_FLOATING_POINT),
            &cdc_task)!=RTEMS_SUCCESSFUL) {
        fatalError("Unable to create the detector task...");
    }
    // start the task
    if (rtems_task_start(cdc_task,cdc_task_main,0)!=RTEMS_SUCCESSFUL) {
        fatalError("Unable to start the detector task...");
    }
    // we are now finished and should kill ourselves
    if (rtems_task_delete(RTEMS_SELF)!=RTEMS_SUCCESSFUL) {
        fatalError("Unable to delete the initialization task...");
    }
    exit(1);
}

// -------------------------------------------------------------------------------------------------
// RTEMS Configuration variables
// -------------------------------------------------------------------------------------------------
// We need the configuration here so that we might adjust the command line arguments according to
// current RTEMS settings
#define CONFIGURE_APPLICATION_NEEDS_CLOCK_DRIVER
#define CONFIGURE_APPLICATION_NEEDS_CONSOLE_DRIVER
#define CONFIGURE_RTEMS_INIT_TASKS_TABLE
#define CONFIGURE_USE_IMFS_AS_BASE_FILESYSTEM
#define CONFIGURE_MICROSECONDS_PER_TICK             1000
#define CONFIGURE_MAXIMUM_TASKS                     2
#define CONFIGURE_MAXIMUM_SEMAPHORES                5
#define CONFIGURE_MAXIMUM_PERIODS                   1
#define CONFIGURE_LIBIO_MAXIMUM_FILE_DESCRIPTORS    6
#define CONFIGURE_INIT_TASK_ATTRIBUTES              RTEMS_LOCAL
#define CONFIGURE_INIT_TASK_INITIAL_MODES          (RTEMS_PREEMPT | \
                                                    RTEMS_NO_TIMESLICE | \
                                                    RTEMS_NO_ASR | \
                                                    RTEMS_INTERRUPT_LEVEL(0))
#define CONFIGURE_INIT

#include <rtems/confdefs.h>

#endif /* NO_RTEMS */
